// FILE: Alarm.c
// COPYRIGHT: (c), Symmetric Research, 2016-2017
//
// Source code for the Alarm.exe custom USBxCH BIN pipeline utility in C.
//
// Open the input and output pipes, then processes USBxCH BIN records as they
// arrive, checking all analog channels for an alarm.  An alarm is triggered if
// a channel amplitude passes the defined threshold.
//
// The alarm is signaled with a printf of ^G ( the "bell" ) which is '\a' in C
// programming.  The BIN record trigger field in which the alarm occurs allows
// downstream pipeline utilities to know about the alarm also.
//
// Two typical alarm applications might include:
//
//   1) monitor a temperature sensor and alarm when it gets too high
//   2) monitor a battery and alarm when its voltage gets too low
//
// The first requires detecting when the temp > THRESHOLD.  The second when
// the voltage < THRESHOLD.
//
// The code below is set up for the > ( greater than ) case.  Change the code
// appropriately for < ( less than ) and recompile as needed.
//
// The file CustomPipeLib.c must also be compiled and linked in to provide the
// pipeline functions used here.
//
// 
//

// Include files:

#include <stdio.h>
#include <stdlib.h>

#include "CustomPipeLib.h"      // for prototypes, BIN, and other SR defines



// Defines:
//
// The alarm threshold is measured in A/D counts.  There are about 10^6 =
// 1,000,000 ( one million ) counts per volt.
//
//

#define THRESHOLD   1000000




// Prototypes:

int DetectAlarm( BIN *pPt );


        

// Main program:

int main( int argc, char *argv[] ) {

        char         *NextProg;
        char         InputPipeName[MAXSTR];
        char         OutputPipeName[MAXSTR];
        SRPIPEHANDLE hInputPipe;
        SRPIPEHANDLE hOutputPipe;
        BIN          Pt;
        int          AlarmValue;


        // Open the report file and announce the program.

        CustomReportOpen( "Alarm" );
        CustomReportPrint( "Alarm starting\n" );



        // Process the command line for the "next program" name.

        if ( argc == 2 )

                NextProg = argv[1];

        else {
                // it is a bad command line, quit ...

                CustomReportPrint( "Usage: Alarm NextProg\n" );
                exit( 0 );
                }


        // Form the input and output pipe names.

        sprintf( InputPipeName,  "%s%s", REQUIRED_PIPE_PREFIX, "Alarm" );
        sprintf( OutputPipeName, "%s%s", REQUIRED_PIPE_PREFIX, NextProg );

    

        // Open the input and output pipes.

        hInputPipe  = CustomOpenInputPipe( InputPipeName );
        hOutputPipe = CustomOpenOutputPipe( OutputPipeName );

          // >> careful programmers might add error checking to confirm the
          // >> input and output pipes were successfully opened ...




        // Loop reading from the input pipe.

        while ( CustomReadInputPipe( hInputPipe, &Pt ) ) {


                // Ok ... a sample point BIN record has been returned by
                //     CustomReadInputPipe() at &Pt ...
                //
                // ... check for an alarm ...

                AlarmValue = DetectAlarm( &Pt );


                // If an alarm occurred, ring the ASCII bell, and set the BIN
                // sample point trigger flag ...

                if ( AlarmValue != 0 ) {

                        printf( "\a" );  // << ring the ASCII bell ...

                        Pt.Trigger = AlarmValue;  // set the BIN trigger flag ...

                        }


                // Pass the BIN sample point on to output pipe ...

                CustomWriteOutputPipe( hOutputPipe, &Pt );

                }


        // The input pipe has failed ( upstream is probably closing down ),
        // close all gracefully ...

        CustomCloseInputPipe(  hInputPipe );
        CustomCloseOutputPipe( hOutputPipe );

        exit( 0 );

}


//------------------------------------------------------------------------------
// ROUTINE: DetectAlarm
//
// PURPOSE: All analog channels are checked.  If the count value exceeds the
//          THRESHOLD the alarm is set.  The alarm signal is saved as a bit
//          indicating the channel the alarm trigger occurred on.
//
//          If AlarmByte is not zero, an alarm occurred.  The channel it
//          occurred on is indicated by the particular bit ( or bits if alarms
//          occurred on several channels at once ) that is set.
//
//          Change the greater than ( Value > THRESHOLD ) test comparison below
//          to ( Value < THRESHOLD ) to set the alarm when an analog channel is
//          less than the THRESHOLD, as when testing for battery failure etc.
//
//------------------------------------------------------------------------------
int DetectAlarm( BIN *pPt ) {

        int i, Value, AlarmByte;


        AlarmByte = 0x00;

        for ( i = 0 ; i < MAXANALOGCHANNELS ; i++ ) {

                // Set an alarm if the channel passes the THRESHOLD.

                Value = pPt->Analog[i];

                if ( Value > THRESHOLD )

                        AlarmByte |= ( 0x01 << i );

                }


        return( AlarmByte );
}

