// FILE: CustomPipeLib.c (LINUX VERSION)
// COPYRIGHT: (c), Symmetric Research, 2016-2019
//
// Source code utility functions for accessing USBxCH BIN data
// from a SrPipe in C.
//

// Include files:

#include <stdio.h>
#include <stdlib.h>

#include <inttypes.h>           // for PRId64 64 bit printf format
#include <errno.h>              // for errno
#include <sys/stat.h>           // for mkfifo
#include <sys/types.h>          // for mkfifo
#include <time.h>               // for nanosleep, struct tm, gmtime, etc

#include "CustomPipeLib.h"      // for prototypes, BIN, and other SR defines


// Defines and global variables:

FILE *RptFp = NULL;             // report file pointer

char MsgBuffer[MAXSTRBIG];      // for building message string



// Custom Pipe Lib Functions:
//
//

//------------------------------------------------------------------------------
// ROUTINE: CustomReportOpen
// PURPOSE: Open a report file.  The provided name will have SrRpt- as a prefix
//          and .rpt as a suffix so the name will be consistent with the
//          standard USBxCH report files.  It exits on failure.
//------------------------------------------------------------------------------
void CustomReportOpen( char *name ) {

        char RptName[MAXSTR];

        sprintf( RptName, "SrRpt-%s.rpt", name );

        RptFp = fopen( RptName, "a" );
        if ( !RptFp )
                exit( 2 );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomReportPrint
// PURPOSE: Write provided message to the previously opened report file.
//------------------------------------------------------------------------------
void CustomReportPrint( char *msg ) {

        if ( RptFp ) {
                fprintf( RptFp, "%s", msg );
                fflush( RptFp );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CustomReportClose
// PURPOSE: Close the previously opened report file.
//------------------------------------------------------------------------------
void CustomReportClose( void ) {

        if ( RptFp ) {
                fclose( RptFp );
                RptFp = NULL;
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CustomOpenInputPipe
// PURPOSE: Open a named pipe for input using the provided name.  Returns a
//          handle to the pipe for passing to other pipe functions.  If the
//          named pipe can not be opened, SR_BADHANDLE is returned.
//------------------------------------------------------------------------------
SRPIPEHANDLE CustomOpenInputPipe( char *InputPipeName ) {

        SRPIPEHANDLE hInputPipe;
        
        while ( 1 ) {

                // Open the pipe for binary reading.

                hInputPipe = fopen( InputPipeName, "rb" );

                // If successful, leave infinite loop.

                if ( hInputPipe )
                        break;
                
                // Otherwise, wait and try again.

                CustomSleep( 500 );
                }

        sprintf( MsgBuffer, "Opened input pipe %s\n", InputPipeName );
        CustomReportPrint( MsgBuffer );

        return( hInputPipe );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomReadInputPipe
// PURPOSE: Read one BIN point from the named input pipe using the provided
//          pipe handle.  Returns 1 on success and 0 otherwise.  The bytes read
//          in are checked to verify they contain the expected BIN point header.
//------------------------------------------------------------------------------
int CustomReadInputPipe( SRPIPEHANDLE hInputPipe, BIN *Pt ) {

        int NumToRead, NumRead, NumRemain, BytesRead, Err;
        char *BytePtr;


        // Error check.

        if ( hInputPipe == SR_BADHANDLE ) {
                CustomReportPrint( "Input pipe bad - can't read.\n" );
                return( 0 );
                }


        // Initialize variables.

        BytePtr   = (char *)Pt;
        NumToRead = NSIZE_BINDATA;
        NumRead   = 0;
        NumRemain = NumToRead - NumRead;


        // Loop around until all requested bytes have been read.
        
        while ( NumRead < NumToRead ) {

                BytesRead = fread( BytePtr, 1, NumRemain, hInputPipe );
                Err       = errno;

                if ( feof( hInputPipe ) )
                        CustomReportPrint( "End of file when reading\n" );
                if ( ferror( hInputPipe ) )
                        CustomReportPrint( "Error when reading\n" );

                BytePtr   += BytesRead;
                NumRead   += BytesRead;
                NumRemain -= BytesRead;

                
                if ( NumRemain == 0 )
                        break;


                if ( Err == EPIPE ) {  // broken pipe
                        CustomReportPrint( "Failed to read with broken input pipe\n" );
                        return( 0 );
                        }

                if ( Err == ENOENT ) { // other end closed
                        CustomReportPrint( "Failed to read input pipe\n" );
                        return( 0 );
                        }

                } // end while


        // Verify we've got a BIN structure by checking its id.

        if ( Pt->Id == BIN_ID )
                return( 1 ); // success

        else {
                CustomReportPrint( "Invalid BIN structure read\n" );
                return( 0 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CustomCloseInputPipe
// PURPOSE: Close the named input pipe using the provided pipe handle.
//------------------------------------------------------------------------------
void CustomCloseInputPipe( SRPIPEHANDLE hInputPipe ) {

        if ( hInputPipe != SR_BADHANDLE ) { 
                fclose( hInputPipe );
                hInputPipe = SR_BADHANDLE;
                CustomReportPrint( "Closed input pipe\n" );
                }
        else {
                CustomReportPrint( "Input pipe already closed\n" );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CustomOpenOutputPipe
// PURPOSE: Open a named pipe for output using the provided name.  Returns a
//          handle to the pipe for passing to other pipe functions.  If the
//          named pipe can not be opened, SR_BADHANDLE is returned.  This
//          function will wait forever for a pipe reader to connect before
//          returning.
//------------------------------------------------------------------------------
SRPIPEHANDLE CustomOpenOutputPipe( char *OutputPipeName ) {
        
        int          mode;
        SRPIPEHANDLE hOutputPipe;
        

        // Make it read write for all.

        mode = S_IRUSR | S_IWUSR |      // user  RW
               S_IRGRP | S_IWGRP |      // group RW
               S_IROTH | S_IWOTH;       // other RW


        // Make fifo.

        if ( mkfifo( OutputPipeName, mode ) != 0 )
                CustomErrorExit( "Failed to make output pipe fifo.\n" );


        // Open pipe.

        hOutputPipe = fopen( OutputPipeName, "wb" );
        if ( !hOutputPipe )
                        CustomErrorExit( "Failed to open output pipe.\n" );

        sprintf( MsgBuffer, "Opened output pipe %s\n", OutputPipeName );
        CustomReportPrint( MsgBuffer );

        return( hOutputPipe );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomWriteOutputPipe
// PURPOSE: Write one BIN point to the named output pipe using the provided
//          pipe handle.  Returns 1 on success and 0 otherwise.
//------------------------------------------------------------------------------
int CustomWriteOutputPipe( SRPIPEHANDLE hOutputPipe, BIN *Pt ) {

        int  NumToWrite, NumWritten, NumRemain, BytesWritten, Err;
        char *BytePtr;


        // Error check.

        if ( hOutputPipe == SR_BADHANDLE ) {
                CustomReportPrint( "Output pipe bad - can't write.\n" );
                return( 0 );
                }


        // Initialize variables.
        
        BytePtr      = (char *)Pt;
        NumToWrite   = NSIZE_BINDATA;
        NumWritten   = 0;
        NumRemain    = NumToWrite - NumWritten;


        // Loop around until all requested bytes have been written.

        while ( NumWritten < NumToWrite ) {

                BytesWritten = fwrite( BytePtr, 1, NumRemain, hOutputPipe );
                Err          = errno;

                BytePtr    += BytesWritten;
                NumWritten += BytesWritten;
                NumRemain  -= BytesWritten;

                if ( NumRemain == 0 )
                        break;
                
                if ( Err == EPIPE ) {
                        CustomReportPrint( "Can't write, broken pipe\n" );
                        return( 0 );
                        }


                if ( Err == ENOENT ) {
                        CustomReportPrint( "Can't write, other end closed\n" );
                        return( 0 );
                        }

                } // end while

        fflush( hOutputPipe );
        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomCloseOutputPipe
// PURPOSE: Close the named output pipe using the provided pipe handle.
//------------------------------------------------------------------------------
void CustomCloseOutputPipe( SRPIPEHANDLE hOutputPipe ) {

        if ( hOutputPipe != SR_BADHANDLE ) { 
                fclose( hOutputPipe );
                hOutputPipe = SR_BADHANDLE;
                CustomReportPrint( "Closed output pipe\n" );
                }
        else {
                CustomReportPrint( "Output pipe already closed\n" );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: CustomErrorExit
// PURPOSE: Writes the provided message to the previously opened report file,
//          closes the report file, and exits.
//------------------------------------------------------------------------------
void CustomErrorExit( char *msg ) {

        CustomReportPrint( msg );
        CustomReportClose( );
        exit( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomSleep
// PURPOSE: Sleeps for at least the provided number of milliseconds.
//------------------------------------------------------------------------------
void CustomSleep( int ms ) {

        struct timespec twait, tleft;
        
        // Linux nanosleep takes structure with seconds and nanoseconds.
        
        tleft.tv_sec  = tleft.tv_nsec = 0;
        twait.tv_sec  = (long)(ms / 1000);
        twait.tv_nsec = (ms - (twait.tv_sec*1000)) * 1000000;
        
        nanosleep( &twait, &tleft );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomPrint64
// PURPOSE: This converts a 64-bit integer into a string.  It returns a pointer
//          to a static string buffer.  When printing TWO strings, copy the
//          first to another buffer before calling this function for the second
//          value !
//------------------------------------------------------------------------------
char *CustomPrint64( SRINT64 Value ) {

// This is a helper routine to convert a 64-bit integer into a string.  It 
// returns a pointer to a static string buffer.  When printing TWO strings,
// copy the first to another buffer before calling this function for the
// second value !

        static char StringBuffer[MAXSTR];

// In 64-bit Linux, longs are 8 bytes and a 64 bit integer is referred to as a
// "long int" which is printed with %ld.
//
// In 32-bit Linux, longs are 4 bytes so a 64 bit number is referred to as a
// "long long int" which is printed with %lld.
//
// The inttypes include accomodates this with the PRId64 define.

        sprintf( StringBuffer, "%" PRId64, Value );

        return( StringBuffer );
}

//------------------------------------------------------------------------------
// ROUTINE: CustomExtractTime
// PURPOSE: This splits a 64 bit time such as the TimeStamp value from a BIN
//          point into human manageable pieces.  The YMDHMS pieces have the 
//          normal values (eg Year is 2019 and January is Month 1).  The 
//          unsigned integer seconds and nanoseconds are counted from Jan 1, 
//          1970.  It is ok to pass NULL for any values you are not interested
//          in.
//------------------------------------------------------------------------------
void CustomExtractTime( SRINT64 TimeStamp,
                        int *Year, int *Month, int *Day,
                        int *Hour, int *Minute, int *Second,
                        unsigned int *IntSec, unsigned int *NanoSec ) {

        unsigned int IntSec1970, NanoSec1970;
        time_t       ttTime;
        struct tm    tmTime;


        // Split 64 bit time into human manageable pieces.
        //
        // Get the seconds since 1970 part as a time_t.
        // And use standard C function to split into YMDHMS.
        // Compute seconds and nanoseconds as unsigned integers.
        // When filling user variables, convert to normal year and month.

        ttTime = (time_t)( TimeStamp / (SRINT64)SR_NSPERSEC );
        tmTime = *gmtime( &ttTime );

        IntSec1970  = (unsigned int)( ttTime );
        NanoSec1970 = (unsigned int)( TimeStamp - 
                                    ((SRINT64)IntSec1970 * (SRINT64)SR_NSPERSEC) );

        if ( Year )     *Year    = tmTime.tm_year + 1900;
        if ( Month )    *Month   = tmTime.tm_mon + 1;
        if ( Day )      *Day     = tmTime.tm_mday;
        if ( Hour )     *Hour    = tmTime.tm_hour;
        if ( Minute )   *Minute  = tmTime.tm_min;
        if ( Second )   *Second  = tmTime.tm_sec;
        if ( IntSec )   *IntSec  = IntSec1970;
        if ( NanoSec )  *NanoSec = NanoSec1970;
}

//------------------------------------------------------------------------------
// ROUTINE: CustomExtractDigital
// PURPOSE: This takes the packed DigitalAll value from a BIN point and extracts
//          the 4 digital inputs into separate integer variables.  It is ok to
//          pass NULL for any values you are not interested in.
//------------------------------------------------------------------------------
void CustomExtractDigital( int DigitalAll,
                           int *Dig0, int *Dig1, int *Dig2, int *Dig3 ) {

        if ( Dig0 )  *Dig0 = (DigitalAll >> 0) & 0x01;
        if ( Dig1 )  *Dig1 = (DigitalAll >> 1) & 0x01;
        if ( Dig2 )  *Dig2 = (DigitalAll >> 2) & 0x01;
        if ( Dig3 )  *Dig3 = (DigitalAll >> 3) & 0x01;
}

//------------------------------------------------------------------------------
// ROUTINE: CustomExtractDegrees
// PURPOSE: This takes the DegC2 (degrees C times 2) value from a BIN point and
//          extracts the true degrees Centigrade and Farenheit.  The degrees C
//          are accurate to 0.5 degrees and the degrees F are accurate to 1
//          degree.  It is ok to pass NULL for any value you are not interested
//          in.
//------------------------------------------------------------------------------
void CustomExtractDegrees( int DegC2, double *DegreesC, double *DegreesF ) {

        int    intF;
        double DegC, DegF;

        // Compute true degrees Centigrade and Farenheit.

        DegC = (double)DegC2 / 2.0;
        DegF = (DegC * 1.8) + 32.0;

        // Round Farenheit degrees to nearest whole number.

        intF = (int)(DegF + 0.5);
        DegF = (double)intF;

        // Fill user variables if passed.
        
        if ( DegreesC )   *DegreesC = DegC;
        if ( DegreesF )   *DegreesF = DegF;
}

//------------------------------------------------------------------------------
// ROUTINE: CustomExtractVoltsGood
// PURPOSE: This takes the packed VoltageGood value from a BIN point and 
//          extracts the power history value.  A value of 1 means power has
//          been good, 0 otherwise.
//------------------------------------------------------------------------------
void CustomExtractVoltsGood( int VoltageGood, int *Vgood ) {

        if ( Vgood )  *Vgood = (VoltageGood >> 1) & 0x0001;
}

