// FILE: TLeader.c
// COPYRIGHT: (c), Symmetric Research, 2016-2017
//
// Source code for the TLeader.exe custom USBxCH BIN pipeline utility in C.
//
// Open the input and output pipes and process USBxCH BIN records as they
// arrive, checking specified analog channels for a trigger event.  A trigger
// occurs when the absoluted amplitude of a channel exceeds a fixed THRESHOLD.
//
// Once a trigger occurs, BIN records have their "trigger flag" marked true
// while the event is underway.  A circular buffer is used to save and mark
// NUMPRE points before the trigger occurred.  The trigger flag continues to be
// marked for NUMPOST points after the trigger.
//
// ALL input pipe BIN records are passed to the output pipe.  This utility only
// marks the BIN record "trigger flag" to reflect the trigger status.  It is up
// to downstream utilities to thin and discard the data if needed.
//
// In the code below, all trigger parameters are DEFINED constants.  If you
// want to change them, assign new values and recompile.  To take trigger
// parameters from the command line or a parameter file, modify the code.  See
// \SR\USBxCH\Exe\Trigger.exe for a utility that already has those features.
//
// The file CustomPipeLib.c must also be compiled and linked in to provide the
// pipeline functions used here.
//
//

// Include files:

#include <stdio.h>
#include <stdlib.h>

#include "CustomPipeLib.h"  // << for SR prototypes, BIN, and other defines



// Trigger parameter defines:
//
// NUMPRE = number of pre-trigger leader points ( BIN records ) to keep
// NUMPOST = number of post-trigger points
// THRESHOLD = count level to signal trigger at
//
// For Scope applications with a quiet geophone, thresholds might be in the
// 4000 range.  At a sample rate of 130 Hz the pre and post values might be 65
// and 130 sample points ( ie 1/2 second before and 1 second after ).
//
// For DVM applications, thresholds might be in the 1000000 (~1v) range with
// pre and post values of 0 and 0 ( ie only show trigger for the actual points
// exceeding the threshold ).
//
// All channels with the "CHECK_CHANNEL" define set to 1 will be scanned for
// setting a trigger.  If you have noisy channel you'd like to ignore, set its
// CHECK define to 0.
//
// Tune these parameters as appropriate for your application ...
//
//

#define NUMPRE                  50
#define NUMPOST                 100
#define THRESHOLD               2500000

#define CHECK_CHANNEL0          1
#define CHECK_CHANNEL1          1
#define CHECK_CHANNEL2          1
#define CHECK_CHANNEL3          1
#define CHECK_CHANNEL4          1
#define CHECK_CHANNEL5          1
#define CHECK_CHANNEL6          1
#define CHECK_CHANNEL7          1




// Tleader.c global variables:

int  ChannelCheck[MAXANALOGCHANNELS] = {

        CHECK_CHANNEL0,  // list of channels to check
        CHECK_CHANNEL1,
        CHECK_CHANNEL2,
        CHECK_CHANNEL3,
        CHECK_CHANNEL4,
        CHECK_CHANNEL5,
        CHECK_CHANNEL6,
        CHECK_CHANNEL7,

        };




// "Pre-leader" circular buffer and indexes:

BIN  CbBinBuffer[ NUMPRE * NSIZE_BINDATA ];

int  CbWriteIndex;      // position to write next value to buffer
int  CbReadIndex;       // position to read next value from buffer



// Prototypes:

void ReportTriggerParameters( void );

void InitCircularBuffer( void );
void CbWritePoint( BIN *pPt );
void CbReadPoint( BIN *pPt );

 int DetectTrigger( BIN *pPt );
void ReportTrigger( BIN *pPt );



// Main program:

int main( int argc, char *argv[] ) {

        char         *NextProg;
        char         InputPipeName[MAXSTR];
        char         OutputPipeName[MAXSTR];
        SRPIPEHANDLE hInputPipe;
        SRPIPEHANDLE hOutputPipe;
        BIN          Pt;
        BIN          LeaderPt;
        int          TriggerOccurred;



        // Open the report file and announce the program.

        CustomReportOpen( "TLeader" );
        CustomReportPrint( "TLeader starting\n" );



        // Process the command line for the "next program" name.

        if ( argc == 2 )

                NextProg = argv[1];

        else { // it is a bad command line, quit ...

                CustomReportPrint( "Usage: TLeader NextProg\n" );
                exit( 0 );
                }



        // Form the input and output pipe names.

        sprintf( InputPipeName,  "%s%s", REQUIRED_PIPE_PREFIX, "TLeader" );
        sprintf( OutputPipeName, "%s%s", REQUIRED_PIPE_PREFIX, NextProg );



        // Open the input and output pipes.

        hInputPipe  = CustomOpenInputPipe( InputPipeName );
        hOutputPipe = CustomOpenOutputPipe( OutputPipeName );

          // >> careful programmers might add error checking to confirm the
          // >> input and output pipes were successfully opened ...



        // Report the trigger parameters.

        ReportTriggerParameters();


        // Initialize the pre-trigger circular buffer.

        InitCircularBuffer();



        // Loop reading from the input pipe.

        while ( CustomReadInputPipe( hInputPipe, &Pt ) ) {


                // An input pipe sample point has been returned ...
                //
                // ... check to see if trigger occured ...

                TriggerOccurred = DetectTrigger( &Pt );



                // Get the first sample point from the leader
                //
                //   ... and mark it if the trigger is on ...
                //

                CbReadPoint( &LeaderPt );

                if ( TriggerOccurred )

                        LeaderPt.Trigger = 1;



                // Send the leader point to the output pipe.
                //
                // ... don't send initialized leader points that are zero ...

                if ( LeaderPt.Id )

                        CustomWriteOutputPipe( hOutputPipe, &LeaderPt );



                // Put the current point into the leader circular buffer.

                CbWritePoint( &Pt );


                // ... and continue on waiting for the input pipe ...

                }



        // The input pipe has failed ( upstream is probably closing down ),
        // close all gracefully ...

        CustomCloseInputPipe(  hInputPipe );
        CustomCloseOutputPipe( hOutputPipe );

        exit( 0 );

}



//------------------------------------------------------------------------------
// ROUTINE: ReportTriggerParameters
//
// PURPOSE: Report the trigger parameters ...
//
//------------------------------------------------------------------------------
void ReportTriggerParameters( void ) {

        int  i;
        char MsgBuffer[MAXSTR];


        // Report parameter values.

        sprintf( MsgBuffer, "NumPre = %d\n", NUMPRE ); CustomReportPrint( MsgBuffer );
        sprintf( MsgBuffer, "NumPost = %d\n", NUMPOST ); CustomReportPrint( MsgBuffer );
        sprintf( MsgBuffer, "Threshold = %d\n", THRESHOLD ); CustomReportPrint( MsgBuffer );

        for ( i = 0 ; i < MAXANALOGCHANNELS ; i++ ) {

                sprintf( MsgBuffer, "Check Channel %d = %d\n", i, ChannelCheck[i] );
                CustomReportPrint( MsgBuffer );

                }

}

//------------------------------------------------------------------------------
// ROUTINE: InitCircularBuffer
// PURPOSE: Initialize circular buffer parameters:
//
//          CbWriteIndex is the index of the next value to be written
//          CbReadIndex  is the index of the next value to be read
//
//------------------------------------------------------------------------------
void InitCircularBuffer( void ) {

        int i;

        for ( i = 0 ; i < NUMPRE ; i++ )

                CbBinBuffer[i].Id = 0;  // <<< zero off the BIN[i] record

                    // you could zero the entire BIN record with:
                    //
                    //   CbBinBuffer[i] = 0
                    //
                    // ... but setting Id = 0 is enough to indicate zero
                    //


        CbWriteIndex = 0;   // start write at the beginning
        CbReadIndex  = 0;   // start read at the beginning

}

//------------------------------------------------------------------------------
// ROUTINE: CbWritePoint
//
// PURPOSE: Write a BIN point to the circular buffer and update the write
//          position index, wrapping if necessary.
//
//------------------------------------------------------------------------------
void CbWritePoint( BIN *pPt ) {


        // Write a BIN record to the circular buffer.

        CbBinBuffer[CbWriteIndex] = *pPt;


        // Update the buffer write position index.

        CbWriteIndex++;

        if ( CbWriteIndex == NUMPRE ) // if off end, wrap
                CbWriteIndex = 0;

}

//------------------------------------------------------------------------------
// ROUTINE: CbReadPoint
//
// PURPOSE: Read a BIN point from the circular buffer and update the read
//          position index, wrapping if necessary.
//
//------------------------------------------------------------------------------
void CbReadPoint( BIN *pPt ) {


        // Read a BIN record from the circular buffer.

        *pPt = CbBinBuffer[CbReadIndex];


        // Update the buffer read position index.

        CbReadIndex++;

        if ( CbReadIndex == NUMPRE ) // if off end, wrap
                CbReadIndex = 0;

}



//------------------------------------------------------------------------------
// ROUTINE: DetectTrigger
//
// PURPOSE: Check for the amplitude ( abs value ) of any analog channel on the
//          check list exceeding the trigger threshold.
//
//------------------------------------------------------------------------------
int DetectTrigger( BIN *pPt ) {

        int i, Amplitude, Triggered;

            static int CountWindow = 0;


        // Check the "active" channels to see if any exceed the threshold.

        Triggered = 0;

        for ( i = 0 ; i < MAXANALOGCHANNELS ; i++ ) {

                if ( ChannelCheck[i] == 1 ) {

                        // Take the abs() value of the analog value.

                        Amplitude = pPt->Analog[i];

                        if ( Amplitude < 0 )
                                Amplitude = -Amplitude;


                        // If greater than threshold, it's a trigger.

                        if ( Amplitude > THRESHOLD )
                                Triggered = 1;

                        }

                }


        // If the trigger occurred, set the trigger window and report it.

        if ( Triggered == 1 ) {

                CountWindow = NUMPRE + 1 + NUMPOST;

                ReportTrigger( pPt );
                }


        // Return 1 if within the trigger window.

        if ( CountWindow > 0 )

                { CountWindow--; return( 1 ); }

        else
                return( 0 ); // no trigger is on ...

}


//------------------------------------------------------------------------------
// ROUTINE: ReportTrigger
//
// PURPOSE: Print a notice showing the sample number of the provided point.
//
//------------------------------------------------------------------------------
void ReportTrigger( BIN *pPt ) {

        char MsgBuffer[MAXSTR];

        sprintf( MsgBuffer,
                 "Trigger detected at sample %s\n",
                 CustomPrint64( pPt->Sample ) );

        CustomReportPrint( MsgBuffer );

}

