// FILE: Last.c
// COPYRIGHT: (c), Symmetric Research, 2019
//
// Source code for the Last.exe custom USBxCH BIN pipeline utility in C.
//
// Open the UsbXch input pipe, then processes UsbXch BIN records as they
// arrive, saving the most recent NBUF BIN records in a circular buffer and
// ASCII output file.
//
// The file CustomPipeLib.c must also be compiled and linked in to provide the
// SR UsbXch pipeline functions used here.
//
//


// Include files:
//
// >>> SEE the include file CustomPipeLib.h FOR THE FOLLOWING >>>
//
//   SR UsbXch BIN record structure layout
//   SR UsbXch BIN pipeline support functions
//

#include <stdio.h>              // for standard C runtime file io
#include <stdlib.h>             // for exit
#include "CustomPipeLib.h"      // for prototypes, BIN structure layout, and other SR defines




// Local function prototypes:

void UpdateBinBuffer( BIN *p );
void PrintBin( BIN *p );



// Local global data: ( ... data that is global to this file only ... )

// Circular buffer for the most recent BIN records.
//
//            NBUF = buffer depth
//     BinBuffer[] = buffer memory
//            ibuf = current location in the circular buffer
//
// To change the buffer depth just change NBUF and recompile ...
//

#define NBUF   10

BIN BinBuffer[NBUF];
int ibuf = 0;


// Output filename and pointer.
//
// To change the output filename just change it here and recompile ...
//
//

char *OutputFileName = "a.txt";
FILE *fpOutput;




// Main program:

int main( int argc, char *argv[] ) {

        SRPIPEHANDLE hInputPipe;
        char         InputPipeName[MAXSTR]; 
        BIN          NewBinRecord;


        // Open the report file and announce the program.

        CustomReportOpen( "Last" );
        CustomReportPrint( "Last starting\n" );



        // Check the command line arguments.

        if ( argc > 1 ) {

                // "Last.exe" does not support any command line arguments.
                //
                //  ... change defines in this source and recompile if needed ...
                //

                CustomReportPrint( "Usage: Last\n" );
                exit( 0 );
                }



        // Form the input pipe name and open it.

        sprintf( InputPipeName,  "%s%s", REQUIRED_PIPE_PREFIX, "Last" );

        hInputPipe  = CustomOpenInputPipe( InputPipeName );

          // >> NOTE: careful programmers might add error checking here to confirm
          // >> the input pipe was successfully opened ...




        // Loop reading and processing BIN records from the input pipe as they arrive.

        while ( CustomReadInputPipe( hInputPipe, &NewBinRecord ) ) {

                // Ok ... a new sample point BIN record has arrived ...

                UpdateBinBuffer( &NewBinRecord );  // add it to the BinBuffer[] ...

                }



        // The input pipe read has failed.
        //
        //   ... most likely the pipeline is closing down ...
        //

        CustomCloseInputPipe(  hInputPipe );  // <<< close all gracefully
        exit( 0 );
}

void UpdateBinBuffer( BIN *pNewBinRecord ) {

        int i, nrec;


        // Open the output file.

        fpOutput = fopen( OutputFileName, "w" );
        
        
        // Save the current BIN record.

        BinBuffer[ibuf] = *pNewBinRecord;
                   ibuf = (ibuf+1)%NBUF;


        // Write out the last NBUF BIN records as ASCII.

        i = ibuf;

        for ( nrec = 0 ; nrec < NBUF ; nrec++ )

                { PrintBin( &BinBuffer[i] );  i = (i+1)%NBUF; }


        // Close the output file.

        fclose( fpOutput );
}

void PrintBin( BIN *pBinRecord ) {
        
        // Convert BIN record fields to ASCII, saving to the output file ...
        //
        // Change the fprintf() format string and add or remove BIN structure
        // record fields as needed.
        //
        // To format the BIN time or temp fields see the functions
        //
        //   CustomExtractTime()      <<< format BIN.TimeStamp
        //   CustomExtractDegrees()   <<< format BIN.DegC2
        //

        fprintf( fpOutput, "%010lu,%10d,%10d,%10d,%10d\n",

               (unsigned long)( pBinRecord->Sample & 0xFFFFFFFFL ),

               pBinRecord->Analog[0],
               pBinRecord->Analog[1],
               pBinRecord->Analog[2],
               pBinRecord->Analog[3]

               );

}

