// FILE: RequestSleep.c (LINUX VERSION)
// COPYRIGHT: (c), Symmetric Research, 2010-2017
//
// RequestSleep is a small helper program that can be easily called from
// a batch file such as SleepWake.bat.  It puts a Windows PC to sleep
// for the specified number of seconds and automatically wakes up and
// exits after those seconds have passed.  It takes two command line
// arguments.  The first specifies the number of seconds to sleep and
// the second specifies whether or not the display should be enabled on
// wakeup.
//

#include <sys/timerfd.h>
#include <time.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>	// for read, close


#define WAKE_TIMER  "WakeTimer"

#define DISPLAY_OFF  0
#define DISPLAY_ON   1

#define USPERSEC     1000000
#define NSPERUS      1000


int SuspendAndWake( int Seconds, int DisplayOn );
int TimerValue( int TimerHandle, int MicroSeconds );

void main( int argc, char **argv ) {

        int Nok, Nsec, Ndisp, ShowUsage;

        // Process command line.

        ShowUsage = 1;
        if ( argc == 3 ) {                                              // if right # args
                Nok = sscanf( argv[1], "%d", &Nsec );                   // get 1st arg
                if ( Nok == 1  &&  Nsec >= 0 ) {                        // if ok
                        Nok = sscanf( argv[2], "%d", &Ndisp );          // get 2nd arg
                        if ( Nok == 1  &&  (Ndisp == 0 || Ndisp == 1) ) // if ok
                                ShowUsage = 0;                          // don't show usage msg
                        }
                }
        if ( ShowUsage == 1 ) {
                printf( "Usage: RequestSleep N D\n\n" );
                printf( "  Where\n" );
                printf( "         N is the number of seconds to sleep before\n" );
                printf( "              automatic wake up.  0 means no wakeup.\n" );
                printf( "         D is the display state on wakeup.  1 for on, 0 for off.\n" );
                exit( 1 );
                }


        // Request timed sleep with auto wake.
        
        if ( SuspendAndWake( Nsec, Ndisp ) )
                exit( 0 );
        else {
                printf( "RequestSleep failed suspend and wake call\n" );
                exit( 1 );
                }
}

//------------------------------------------------------------------------------
// ROUTINE: SuspendAndWake
// PURPOSE: Suspend the system by putting it into standby mode.  Optionally
//          set a timer to wake the system in the specified number of seconds.
//          Also specify whether or not to turn the display on when waking.
//          Returns 1 for success, 0 on error.
//------------------------------------------------------------------------------
int SuspendAndWake( int Seconds, int DisplayOn ) {

        int                TimerHandle, nbytes;
        int                Ok, Result, MicroSeconds;
        unsigned long long missed;      // uint64_t
        
        // If time requested, open and set wake timer.

        if ( Seconds > 0 ) {

                TimerHandle = timerfd_create( CLOCK_MONOTONIC, 0 ); // 0 is flags
                if ( TimerHandle == -1 )
                        return( 0 );

                MicroSeconds = Seconds * USPERSEC;
                
                if ( TimerValue(  TimerHandle, MicroSeconds ) == 0 )
                        return( 0 );
                }



        // Put system into standby.

//        Ok = SetSuspendState( );
        Ok = 0;
        if ( !Ok ) {
                if ( TimerHandle > 0 ) close( TimerHandle );
                return( 0 );
                }
                


        // Wait for the timer.

        if ( Seconds > 0 ) {

		nbytes = read( TimerHandle, &missed, sizeof( missed ) );

                if ( missed > 0 )
                        Result = 0;
                else
                        Result = 1;

                // Indicate that we require the display.

//                if ( DisplayOn  || ( Result == 0 ) )
//                        RestoreDisplay( );

                close( TimerHandle );
                
                if ( Result == 0 )
                        return( 0 );
                }


        return( 1 );
}

//------------------------------------------------------------------------------
// ROUTINE: TimerValue
// PURPOSE: Assigns the wait time in microseconds to the timer associated with 
//          the provided handle.  Returns 1 for success, 0 on error.
//------------------------------------------------------------------------------
int TimerValue( int TimerHandle, int MicroSeconds ) {

        int    fd, ret;
        time_t isec;
        long   nano;
        struct itimerspec itval;

        fd = (int)TimerHandle;

        isec = MicroSeconds / USPERSEC;
        nano = (MicroSeconds - (isec *  USPERSEC) ) * NSPERUS;

        itval.it_interval.tv_sec  = isec;
        itval.it_interval.tv_nsec = nano;
        itval.it_value.tv_sec     = isec;
        itval.it_value.tv_nsec    = nano;

        ret = timerfd_settime( fd, 0, &itval, NULL ); // 0 means relative, NULL is old value


        if ( ret == 0 )         // success
                return( 1 );
        
        else // ret == -1       // error
                return( 0 );
}

