// FILE: sr2sc_plugin.c
// COPYRIGHT: (c), Symmetric Research, 2016-2018
//
// This is an interface utility to connect the SeisComp seismic
// management system with the SR USBxCH data acquisition pipeline
// utilities.
//
// It first gathers any SeisComp parameters set using the graphical
// scconfig program.  Then it runs the Sr2ScPipeline batch file which
// starts a chain of SR programs that produce an output pipe with BIN
// records containing the acquired data in a convenient format.  The
// program chain includes:
//
//        Blast    to acquire data into PAK format
//        Pak2Bin  to convert from PAK to BIN format
//        Interp   to resample to a simple timebase
// 
// The last program in the chain sends its data to this plugin which
// goes into a loop sending data from BIN files output by a SR USBxCH 24
// bit system to Seiscomp as raw3 packets.
//
// When a seiscomp stop or kill signal is received, the plugin shuts
// down which should kill the other programs in the pipeline chain.
// But just to be safe, it also calls a second batchfile to kill the
// pipeline programs.
//
// The most recent version of this plugin can be downloaded for free from
// the symres.com website as it is included as part of the USBxCH software
// in the Special App - SeisComp subdirectory under Examples.
//
//
//
// changes:
//
// Revision 1.0  2018/12/31      W.Tucker:
//       first version
// Revision 0.9  2017/11/21      W.Tucker:
//       beta version
// Revision 0.8  2016/03/12      W.Tucker:
//       preliminary version


#include <stdio.h>
#include <stdlib.h>
#include <time.h>        // for gmtime, struct tm, time_t, etc
#include <fcntl.h>       // for O_CREAT, O_WRONLY, etc
#include <sys/stat.h>    // for _S_IWRITE 
#include <string.h>      // for strcmp
#include <errno.h>       // for errno
#include <unistd.h>      // for access, close function and STDIN_FILENO
#include <signal.h>      // for SIGTERM, struct sigaction, etc
#include <stdarg.h>      // for va_start

#define LOGOFLAGS (O_CREAT|O_EXCL|O_WRONLY)                 // open  flags for log file 
#define LOGSFLAGS (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH) // share flags for log file 

#include "CustomPipeLib.h"      // for prototypes, BIN, and other SR defines
#include "plugin.h"	        // for send_raw3, etc utilities, etc



// Program copyright and screen title:

char *Copyright =

                "\n\nsr2sc_plugin.exe : Copyright (c), Symmetric Research, 2016-2018\n\n";

char *ScreenTitle =

                "USBxCH sr2sc_plugin : (c) SR 2016 : Rev {2018/12/31}\n\n";


// Defines and typedefs.

#define MAXSTR                 256  // Maximum dimension for some strings
#define MAXSTRBIG              512  // Maximum dimension for msg buffer
                                    // Maximum number of channels allowed (16)
#define MAX_SC_CHAN (MAXANALOGCHANNELS + MAXDIGITALCHANNELS + MAXSPECIALCHANNELS)

#define QUALITY_OK            100   // time quality is 100% good
#define QUALITY_QUESTIONABLE   50   // time quality is  50% good, low satellites
#define QUALITY_BAD             0   // time quality is   0% good

#define MIN_GOOD_TIME   1454695792  // minimum seconds to consider time is ok

#define MAX_TRACEBUF_PTS 10000      // 10kHz
#define MAX_TRACEBUF_SIZ (MAX_TRACEBUF_PTS*sizeof(int))


#define NSPERSEC   1000000000       // nanoseconds / second
#define USPERSEC      1000000       // microseconds / second
#define NSPERUS          1000       // nanoseconds / microseconds


// Global variables:

char SeedlinkDir[MAXSTR];
char YmdHmsDir[MAXSTR];
char DebugRpt[MAXSTR];
char PipelineCmd[MAXSTR];

//DBG
FILE *fp;


// Bin2Sc RUN TIME PARAMETERS:
//
// Both the user specified command line parameters and fixed hardwired
// parameters are allocated here.  Access this structure with statements
// like: PP.ProgramName.
//
//

struct Bin2ScParametersLayout {

	char  *SectionName;
	int    fStopExecution;
	int    Verbosity;                 // 0,1,2,3 = few, some, many... messages
	int    CustomReportReady;	  // is SR report file available

        int    OutputRate;                // output sampling rate in Hz
        int    TimeQualityCheck;          // Check GPS time quality or not

	int    NumReadySamples;           // Number of samples in current data dump
	int    MaxReadySamples;           // Number of samples in one second
        int    NumChannelEnabled;         // Number of channels enabled

	char   StationName[MAXSTR];
	char   ChannelName[MAX_SC_CHAN][MAXSTR];
	int    ChannelEnable[MAX_SC_CHAN];

	char   LogFileName[MAXSTR];
	char   RptFileName[MAXSTR];
	char   ProgramName[MAXSTR];

	SRPIPEHANDLE hInputPipe;

	char   InputPipeName[MAXSTR];

	char  *TraceBuffer[MAX_SC_CHAN];  // trace buffer
	int   *TracePt[MAX_SC_CHAN];      // where in trace buffer the current point is


} ProgramParameters = {


	// default parameter values:
	//

	"Bin2Sc",            // SectionName, do not change
	0,                   // fStopExecution, 0 -> no (default), 1 -> yes
	0,                   // Verbosity, 0 -> no debugging messages
	0,	             // CustomReportReady, 0 -> no, not open yet

	100,                 // OutputRate, 100 Hz
        0,                   // 0 -> assume time good, don't check quality 1 -> do

	0,                   // NumReadySamples, 0
	100,                 // MaxReadySamples = PP.OutputRate
        0,                   // NumChannelEnabled, starts at 0


	// all trailing filename and structure areas are
	// initialized to 0 by the compiler ...

};

#define PP      ProgramParameters  // PP = shorthand for decimate program parameter struct


// Function Prototypes:

 int main( int argc, char **argv );
void PrepareYmdHms( char *ParentDir );
void RunPipelineCmd( char *PipeCmd );
void Initialize( int argc, char **argv );
void SkipFileLines( FILE *fp, int NumLines );
void FormatRunTimeSummary( void );
void OutputPoint( BIN *Pt );
void SendTraces( SRINT64 StartTime, int TimeQuality );
void ScCheckTerminate( void );
void CleanExit( int ExitCode );
void SignalInit( void );
void SignalHandle( int x );
void LogMessage( char *msg, int level, int err );




//------------------------------------------------------------------------------
// ROUTINE: main()
// PURPOSE: Main entry point for sr2sc_plugin.  This program provides a 
//          connection between Symmetric Research 24 bit USBxCH A/D (DAQ) 
//          systems and the SeisComp program maintained by GFZ.
//
//          It creates and changes to a YMDHMS directory, then calls the
//          Sr2ScPipeline script to start an SR acquisition pipeline.  Next it
//          goes into a loop behaving like a standard SR BIN pipeline utility
//          reading BIN records and processing them.  The processing is using
//          send_raw3 to send the data on to SeisComp.  It continues until a
//          SeisComp terminate request is received, a Ctrl+C is pressed, or a
//          fatal error such as a broken pipe further up the pipeline occurs.
//------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {

	BIN   Pt;
	char  MsgBuffer[MAXSTRBIG];
	int   i;

	PP.Verbosity         = 6;
	PP.CustomReportReady = 0;
	sprintf( SeedlinkDir, "/home/sysop/seiscomp3/var/lib/seedlink" );

	

	// Set up debugging report file.

	sprintf( DebugRpt, "%s/Dbg.rpt", SeedlinkDir );
	fp = fopen( DebugRpt, "w" );
	if ( !fp )
		printf( "sr2sc_plugin can't open debugging report file\n" );
	else {
		fprintf( fp, "sr2sc_plugin called with argc = %d\n", argc );
		for ( i = 0 ; i < argc ; i++ ) {
			fprintf( fp, "arg[%d] = %s\n", i, argv[i] );
			}
		fflush( fp );
		}


	// Set up signals.

	SignalInit( );



	// Make, then change to YMDHMS directory.

	PrepareYmdHms( SeedlinkDir );



	// Open the report file and announce the program.

	PP.CustomReportReady = 1;
	CustomReportOpen( "Bin2Sc" );

	LogMessage( "Bin2Sc (sr2sc_plugin) starting\n", 1, 0 );


        // Initialize variables.  We get some from the command line
        // and some from the plugins.ini file.  Users can use scconfig
        // to enter values into the plugins.ini file.  Also allocate
        // space for trace buffers.

	Initialize( argc, argv );



	// Print out the final parameter summary before running.

	FormatRunTimeSummary( );



	// Start SR to SC pipeline utility chain.

	sprintf( PipelineCmd, "%s/Sr2ScPipeline", SeedlinkDir );
	RunPipelineCmd( PipelineCmd );
	

	
	// Form the name and open the input pipe.

	sprintf( PP.InputPipeName,  "%s/%s%s", YmdHmsDir, REQUIRED_PIPE_PREFIX, "Bin2Sc" );
	PP.hInputPipe = CustomOpenInputPipe( PP.InputPipeName );

	  // >> careful programmers might add error checking to confirm the
	  // >> input pipes was successfully opened ...



	// While there is data...
	// Infinite loop reading BIN packets from the input pipe and
	// sending raw3 data to Seiscomp.  Exits if the other end of the
	// input pipe is closed or a stop signal is received.

	LogMessage( "Begin sending to SeisComp ...\n", 0, 0 );
	
	while ( CustomReadInputPipe( PP.hInputPipe, &Pt ) ) {

		ScCheckTerminate( );	// Terminates if stop signal received

		OutputPoint( &Pt );	// Output to SeisComp
		}

	LogMessage( "Stopping Bin2Sc (sr2sc_plugin).\n\n", 0, 0 );



	// The input pipe has failed ( upstream is probably closing down ),
	// announce final results and close all gracefully ...

	sprintf( MsgBuffer, "SUCCESS: Final Bin2Sc (sr2sc_plugin) data >> SeisComp\n\n" );
	LogMessage( MsgBuffer, 0, 0 );


	CleanExit( 0 );


	// Return from main:
	//
	// Linux *requires* main to be of type integer with return 0 for
	// success.  Windows doesn't care ...
	//
	//

        exit( 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: PrepareYmdHms
// PURPOSE: Get the current date and time as YMDHMS, make a directory under the
//          supplied parent directory with that info.  Then change the working 
//          directory of this executable to this newly created directory.
//	    Because seiscomp started this plugin, the original working directory
//	    is the directory where seiscomp itself was started (typically
//	    /home/sysop/seiscomp3).  Assumes a valid parent directory is given.
//------------------------------------------------------------------------------
void PrepareYmdHms( char *ParentDir ) {

	time_t UtcTimeInSecondsSince_1970;
	struct tm *LocTime;
	int Permissions, SysRet, err;
	char MsgBuffer[MAXSTRBIG];


	// Get the current time.

	time( &UtcTimeInSecondsSince_1970 );
	LocTime = localtime( &UtcTimeInSecondsSince_1970 );



	// Prepare the directory name: YYYY-MM-DD-at-HH-MM-SS
	//
	//    YMDHMS = ( year, month, day, hours, minutes, seconds)
	//
	//

	sprintf( YmdHmsDir, "%s/%04d-%02d-%02d-at-%02d-%02d-%02d",

		 ParentDir,
		 LocTime->tm_year+1900,
		 LocTime->tm_mon+1,
		 LocTime->tm_mday,
		 LocTime->tm_hour,
		 LocTime->tm_min,
		 LocTime->tm_sec

	       );

	LogMessage( "YMDHMS directory is ...\n", 0, 0 );
	LogMessage( YmdHmsDir, 0, 0 );
	LogMessage( "\n", 0, 0 );


	Permissions = (S_IRWXU | S_IRWXG | S_IRWXO);  // Read,Write,Execute
					              // for User,Group,Other


	// Make the directory.

	SysRet = mkdir( YmdHmsDir, Permissions );
	if ( SysRet == 0 ) {
		sprintf( MsgBuffer, "Made directory %s\n", YmdHmsDir );
		LogMessage( MsgBuffer, 1, 0 );
		}
	else {
		err = errno;
		sprintf( MsgBuffer, "Failed to make directory %s, error %d\n",
			 YmdHmsDir, err );
		LogMessage( MsgBuffer, 0, 1 );
		}


	// Change to it.

	SysRet = chdir( YmdHmsDir );
	if ( SysRet == 0 ) {
		sprintf( MsgBuffer, "Change working directory to %s\n", YmdHmsDir );
		LogMessage( MsgBuffer, 1, 0 );
		}
	else {
		err = errno;
		sprintf( MsgBuffer, "Failed to change working dir to %s, error %d\n",
			 YmdHmsDir, err );
		LogMessage( MsgBuffer, 0, 1 );
		}
}

//------------------------------------------------------------------------------
// ROUTINE: RunPipelineCmd()
// PURPOSE: Run the specified SR to SC pipeline script.  Users may edit the
//          pipeline startup script file as needed, but the last pipeline 
//          utility should ALWAYS have Bin2Sc as its output location.  Exits on
//          failure.
//------------------------------------------------------------------------------
void RunPipelineCmd( char *PipeCmd ) {

	int SysRet;
	char MsgBuffer[MAXSTRBIG];

	SysRet = system( PipeCmd );
	if ( SysRet == 0 ) {
		sprintf( MsgBuffer, "sr2sc_plugin running pipe script \n\t\t\t%s ...\n",
			 PipeCmd );
		LogMessage( MsgBuffer, 1, 0 );
		}
	else {
		sprintf( MsgBuffer, "check report files, sr2sc_plugin failed to run pipe script \n\t\t\t%s\n",
		       PipeCmd );
		LogMessage( MsgBuffer, 0, 1 );
		}
}

//------------------------------------------------------------------------------
// ROUTINE: Initialize
// PURPOSE: Hardwire all parameters.  Later, get from cfg file instead.
//          Also allocate space for trace buffers.
//------------------------------------------------------------------------------
void Initialize( int argc, char **argv ) {

        int  i, ichan, optind;
        char MsgBuffer[MAXSTRBIG], IniFileName[MAXSTR];
        char *LinePtr, Line[MAXSTR], SectionLine[MAXSTR];
        FILE *fpini;
        

	LogMessage( "Initialize starting ...\n", 1, 0 );


        // Set variable defaults.
        
	sprintf( PP.StationName, "%s", "SRXX" );		// name of this station

        for ( ichan = 0; ichan < MAX_SC_CHAN ; ichan++ ) {

		PP.ChannelEnable[ichan] = 0;			// disabled
                sprintf( PP.ChannelName[ichan], "C%X", ichan );
                }

	PP.ChannelEnable[0] = 1;	// enabled
	PP.Verbosity        = 6;	// amount of info sent to log file 0-7

	// *** MUST MATCH [Interp] values ***

	PP.OutputRate      = 200;		// sampling rate
        PP.MaxReadySamples = PP.OutputRate;



        // Process command line arguments.  The command line is provided 
        // in the seedlink_plugin.tpl template and is used to pass seedlink
        // parameters set in the profile_XXX files using scconfig.
        

        // Show the arguments.

	sprintf( MsgBuffer, "Number of command line arguments is %d\n", argc );
	LogMessage( MsgBuffer, 3, 0 );

	for ( i = 0 ; i < argc ; i++ ) {
		sprintf( MsgBuffer, "Arg[%d] is %s\n", i, argv[i] );
		LogMessage( MsgBuffer, 3, 0 );
		}

        // Process the arguments.

	for ( optind = 2 ; optind < argc ; optind++ ) {
		if ( strcmp( argv[optind], "-v" ) == 0) {
			PP.Verbosity = atoi( argv[++optind] );
                        }
                // Add additional arguments here ...
                }

        LogMessage( "After process seedlink plugin tpl command line:\n", 5, 0 );
        sprintf( MsgBuffer, "verbosity = %d\n", PP.Verbosity );
	LogMessage( MsgBuffer, 5, 0 );



        // Process the plugins.ini file.  It is generated by combining the
        // plugins.ini.tpl template with parameters set in the profile_XXX
        // files using scconfig.  Here, we just do very simple fscanf parsing.
        // So the number and order of lines in the template MUST agree with
        // this code.
        //
        // Note: Don't allow any blank lines in the ini once you start
        // using fscanf.


        // Open plugins.ini file.

        sprintf( IniFileName, "%s/plugins.ini", SeedlinkDir );
        fpini = fopen( IniFileName, "r" );
        if ( !fpini ) {
                sprintf( MsgBuffer, "Could not open ini file %s\n", IniFileName );
                LogMessage( MsgBuffer, 0, 1 );
                }


        // Skip first 3 lines, 2 comment and one blank.
        
        SkipFileLines( fpini, 3 );



        // Now get and verify the section line.

        sprintf( SectionLine, "[%s]\n", argv[1] );

        LinePtr = fgets( Line, MAXSTR-1, fpini );
        if ( strcmp( Line, SectionLine ) != 0 ) {
                sprintf( MsgBuffer, "Plugins ini file not for section %s\n", argv[1] );
                LogMessage( MsgBuffer, 0, 1 );
                }


        // Skip 6 lines, 4 comment and 2 blank.

        SkipFileLines( fpini, 6 );


        // Now get the two parameters - station, and samplerate.

        LinePtr = fgets( Line, MAXSTR-1, fpini );
        if ( sscanf( Line, "station = %s\n", PP.StationName ) != 1 ) {
                sprintf( MsgBuffer, "Plugins ini bad format on line 11, expected station\n" );
                LogMessage( MsgBuffer, 0, 1 );
                }

        LinePtr = fgets( Line, MAXSTR-1, fpini );
        if ( sscanf( Line, "samplerate = %d\n", &(PP.OutputRate) ) != 1 ) {
                sprintf( MsgBuffer, "Plugins ini bad format on line 12, expected samplerate\n" );
                LogMessage( MsgBuffer, 0, 1 );
                }
        PP.MaxReadySamples = PP.OutputRate;



        // Skip 1 blank line.

        SkipFileLines( fpini, 1 );



        // Now get the channel enable/disable values. 1=enable, 0=disable.
        // First create the string to match, then match it.

        for ( ichan = 0 ; ichan < 8 ; ichan++ ) {

                LinePtr = fgets( Line, MAXSTR-1, fpini );
                if ( sscanf( Line, "c%1denable = %d\n", &i, &(PP.ChannelEnable[ichan]) ) != 2 ) {
                        sprintf( MsgBuffer, "Plugins ini bad format, expected c%1denable\n", ichan );
                        LogMessage( MsgBuffer, 0, 1 );
                        }

                sprintf( MsgBuffer, "Channel found i=%d, looking for ichan=%d, Enable=%d\n",
                         i, ichan, PP.ChannelEnable[ichan] );
                LogMessage( MsgBuffer, 3, 0 );

                if ( i != ichan ) {
                        sprintf( MsgBuffer, "Plugins ini bad format, c%1denable out of order\n", i );
                        LogMessage( MsgBuffer, 0, 1 );
                        }

                
                } // end for ichan < 8
        

        // Skip 1 blank line.

        SkipFileLines( fpini, 1 );


        // Now get the time quality parameter.

        LinePtr = fgets( Line, MAXSTR-1, fpini );
        if ( sscanf( Line, "timequalitycheck = %d\n", &(PP.TimeQualityCheck) ) != 1 ) {
                sprintf( MsgBuffer, "Plugins ini bad format on line 23, expected timequalitycheck\n" );
                LogMessage( MsgBuffer, 0, 1 );
                }



        // Close ini file.

        fclose( fpini );
        

        LogMessage( "After process plugins ini tpl file:\n", 5, 0 );
        sprintf( MsgBuffer, "station = %s, rate = %d, timequalitycheck = %d\n",
                 PP.StationName, PP.OutputRate, PP.TimeQualityCheck );
	LogMessage( MsgBuffer, 5, 0 );



        // Count up number of enabled channels.

        PP.NumChannelEnabled = 0;
        
        for ( ichan = 0; ichan < MAX_SC_CHAN ; ichan++ ) {

		if ( PP.ChannelEnable[ichan] == 1 )
                        PP.NumChannelEnabled++;
                }
        

	// Allocate space for the output trace buffers including both header + data.

	for ( ichan = 0 ; ichan < MAX_SC_CHAN ; ichan++ ) {

		sprintf( MsgBuffer, "Allocating space for channel %2d of %2d\n", ichan, MAX_SC_CHAN );
		LogMessage( MsgBuffer, 5, 0 );

		PP.TraceBuffer[ichan] = (char *) malloc( MAX_TRACEBUF_SIZ ); // alloc max size

		if ( PP.TraceBuffer[ichan] == NULL )
			LogMessage( "Cannot allocate all trace buffers\n", 0, 1 );

		// Set up pointers to the data.

		PP.TracePt[ichan]  = (int *)(PP.TraceBuffer[ichan]);
		}
	
	// Verify ???



        
#ifdef NOT_YET

        // This could be an alternate way of processing command line arguments
	
	while((c = getopt(argc, argv, "vDf:Vh")) != EOF)
	{
		switch(c)
		{
			case 'v': ++verbosity; break;
			case 'X': verbosity = atoi(optarg); break;
			case 'D': daemon_mode = true; break;
			case 'f': config_file = optarg; break;
			case 'V': cout << ident_str << endl;
				  exit(0);
			case 'h': fprintf(stdout, help_message, get_progname(argv[0]).c_str());
				  exit(0);
			case '?': fprintf(stderr, opterr_message, get_progname(argv[0]).c_str());
				  exit(1);
		}
	}
	
#endif // NOT_YET


	LogMessage( "Initialize ending ...\n", 1, 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: SkipFileLines
// PURPOSE: To skip lines in the provided file.  Assumes file is already open.
//------------------------------------------------------------------------------
void SkipFileLines( FILE *fp, int NumLines ) {

        int  i;
        char *LinePtr, Line[MAXSTR];
	char MsgBuffer[MAXSTRBIG];


        for ( i = 0 ; i < NumLines ; i++ ) {
        
                LinePtr = fgets( Line, MAXSTR-1, fp );
                sprintf( MsgBuffer, "Skipping: %s\n", LinePtr );
                LogMessage( MsgBuffer, 5, 0 );
                }

}

//------------------------------------------------------------------------------
// ROUTINE: FormatRunTimeSummary
// PURPOSE: Do the actual formatting and printing of the run time summary info.
//------------------------------------------------------------------------------
void FormatRunTimeSummary( void ) {

        int  i;
	char MsgBuffer[MAXSTRBIG];
 static char *OnOff[2] = { "Off", "On" };

	 LogMessage( "FormatRunTimeSummary starting ...\n", 2, 0 );
 
        // Print out filled values.

        LogMessage( ScreenTitle, 0, 0 );        // announce program
	LogMessage( "\n", 0, 0 );
	
        sprintf( MsgBuffer, "  %s runtime summary ... \n\n", PP.SectionName ); LogMessage( MsgBuffer, 0, 0 );
        sprintf( MsgBuffer, "        Verbosity:  %d\n",      PP.Verbosity   ); LogMessage( MsgBuffer, 0, 0 );
        sprintf( MsgBuffer, "     Station Name:  %s\n",      PP.StationName ); LogMessage( MsgBuffer, 0, 0 );
        sprintf( MsgBuffer, "    Sampling Rate:  %d\n",      PP.OutputRate );  LogMessage( MsgBuffer, 0, 0 );
        sprintf( MsgBuffer, " TimeQualityCheck:  %s\n",OnOff[ PP.TimeQualityCheck ] ); LogMessage( MsgBuffer, 0, 0 );
        sprintf( MsgBuffer, "\n" );

        for ( i = 0 ; i < 10 ; i++ ) {
		sprintf( MsgBuffer, "         Channel%1d:   %s %s\n", i,
			 PP.ChannelName[i], OnOff[ PP.ChannelEnable[i] ] );    LogMessage( MsgBuffer, 0, 0 );
	}
        for ( i = 10 ; i < 16 ; i++ ) {
		sprintf( MsgBuffer, "         Channel%2d:  %s %s\n", i,
			 PP.ChannelName[i], OnOff[ PP.ChannelEnable[i] ] );    LogMessage( MsgBuffer, 0, 0 );
	}
	LogMessage( "\n\n\n", 0, 0 );

	LogMessage( "FormatRunTimeSummary ending ...\n", 2, 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: OutputPoint
// PURPOSE: Put BIN point into trace buffer and send to SeisComp when ready.
//------------------------------------------------------------------------------
void OutputPoint( BIN *Pt ) {

 static int     FirstTimeBad  = 1;
 static int     FirstTimeGood = 1;
 static int     LastSec       = 0;
 static int     TimeQuality   = QUALITY_OK;  // Data timing quality, ok or questionable
 static int     LastQuality   = QUALITY_BAD; // Previous data timing quality
 static SRINT64 StartTime     = SR_ZERO64;
 
        int     ichan, idig, powerH, tempC10, thisSec;
        char    MsgBuffer[MAXSTRBIG];

	LogMessage( "OutputPoint starting ...\n", 6, 0 );


        // Don't send data until GPS time is good.

        thisSec = (int)((Pt->TimeStamp) / ((SRINT64)(NSPERSEC)));

        if ( PP.Verbosity >= 6 ) {
                sprintf( MsgBuffer, "TimeStamp = %s, thisSec = %d\n",
                         CustomPrint64( Pt->TimeStamp ), thisSec );
                LogMessage( MsgBuffer, 6, 0 );
                }


        if ( thisSec < MIN_GOOD_TIME ) {
                if ( FirstTimeBad ) {
                        LogMessage( "GPS time not ready yet, no data sent\n", 0, 0 );
                        FirstTimeBad = 0;
                        }
                return;
                }
        else {
                if ( FirstTimeGood ) {
                        LogMessage( "GPS time now available, sending data soon\n", 0, 0 );
                        FirstTimeGood = 0;
                        LastSec       = thisSec;
                        }
                }


        // For first point in trace buffer, update trace data pointers,
        // start time, and quality flag.

        if ( PP.NumReadySamples == 0 ) {

                if ( PP.Verbosity >= 7 ) {
                        sprintf( MsgBuffer,
                                 "First point in tracebuf for sec %d\n",
                                 thisSec );
                        LogMessage( MsgBuffer, 7, 0 );
                        }

                TimeQuality = QUALITY_OK;
                StartTime   = Pt->TimeStamp;

                
                // Prepare pointers for start of each channel in combined buffer.
                
                for ( ichan = 0 ; ichan < MAX_SC_CHAN ; ichan++ )
                        PP.TracePt[ichan]  = (int *)(PP.TraceBuffer[ichan]);

                } // end if PP.NumReadySamples == first point


        // Save current point: analog, digital, & special channels.

        // Fill in this point for analog channels.

	LogMessage( "Ready to fill analog channels\n", 6, 0 );

	for ( ichan = 0 ; ichan < MAXANALOGCHANNELS ; ichan++ ) {
                *(PP.TracePt[ichan]) = Pt->Analog[ichan];
                }


        // Fill in this point for 4 digital channels.

	LogMessage( "Ready to fill digital channels\n", 6, 0 );
        for ( idig = 0 ; idig < MAXDIGITALCHANNELS ; idig++, ichan++ ) {
                *(PP.TracePt[ichan]) = ((Pt->DigitalAll >> idig) & 0x01);
                }


        // Fill in this point for 4 special channels.
        
	LogMessage( "Ready to fill special channels\n", 6, 0 );
        tempC10 = Pt->DegC2 * 5;                    // degree C x 2 x 5 = 10 x degrees C
        powerH  = (Pt->VoltageGood & 0x02) ? 1 : 0; // Is power history bit good ?
//      powerH  = SrUsbXchIsPowerBitGood( Pt->VoltageGood, USBXCH_POWER_HISTORY );

        *(PP.TracePt[ichan++]) = Pt->PpsToggle;
        *(PP.TracePt[ichan++]) = powerH;
        *(PP.TracePt[ichan++]) = tempC10;
        *(PP.TracePt[ichan++]) = Pt->NumSat;


        // Count this point.
        
        PP.NumReadySamples++;

        if ( PP.Verbosity >= 6 ) {
                sprintf( MsgBuffer, "Point %d at time %s\n",
                         PP.NumReadySamples, CustomPrint64(Pt->TimeStamp) );
		LogMessage( MsgBuffer, 6, 0 );
                }



        // Check timing quality if requested, otherwise assumes it is good.

        if ( PP.TimeQualityCheck == 1 ) {
                
                if ( Pt->NumSat > 3 ) { // GPS time is good

                        TimeQuality = QUALITY_OK;
                        if ( (TimeQuality != LastQuality) && (PP.Verbosity >= 4) ) {
                                sprintf( MsgBuffer,
                                         "Time quality ok (Nsat = %d)\n",
                                         Pt->NumSat );
                                LogMessage( MsgBuffer, 4, 0 );
                                }
                        }

                else { // NumSat <= 3   // not enough satellites for good time
                        
                        TimeQuality = QUALITY_QUESTIONABLE;
                        if ( (TimeQuality != LastQuality) && (PP.Verbosity >= 4) ) {
                                sprintf( MsgBuffer,
                                         "Time quality medium (Nsat = %d)\n",
                                         Pt->NumSat );
                                LogMessage( MsgBuffer, 4, 0 );
                                }
                        } // end NumSat < 3

                } // end if TimeQualityCheck

        LastQuality = TimeQuality;
        

        // If the trace buffer is full, send the data.

        if ( PP.NumReadySamples == PP.MaxReadySamples ) {
                
                if ( PP.Verbosity >= 6 )  {
                        sprintf( MsgBuffer,
                                 "Sending max number of samples %d (= %d) in tracebuf for sec %d, time %s\n",
                                 PP.NumReadySamples, PP.MaxReadySamples,
                                 thisSec, CustomPrint64( Pt->TimeStamp ) );
                        LogMessage( MsgBuffer, 6, 0 );
                        }
                
                SendTraces( StartTime, TimeQuality );

                } // end if PP.NumReadySamples == last point


        // Else, normal point

        else {

                // Prepare for next point.

                for ( ichan = 0 ; ichan < MAX_SC_CHAN ; ichan++ )
                        (PP.TracePt[ichan])++;

                } // end else (normal point)


	LogMessage( "OutputPoint ending ...\n", 6, 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: SendTraces
// PURPOSE: Send traces saved in the trace buffer arrays to SeisComp.  The data 
//          is sent one channel at a time.
//------------------------------------------------------------------------------
void SendTraces( SRINT64 StartTime, int TimeQuality ) {

        int          ichan, microsecond, r, intsec1970, nanosec1970;
        struct ptime seedlinktime;
        struct tm   *ptm;
	time_t       ttTime;
	double	     StartTimeSeconds, usec;
	char         MsgBuffer[MAXSTRBIG];
 static int          FirstTimeMsg = 0;


        LogMessage( "SendTraces starting ...\n", 6, 0 );

	
        // Set up seedlink time for start of this set of raw data.

        intsec1970  = (unsigned int)(StartTime / (SRINT64)NSPERSEC);
        nanosec1970 = (unsigned int)(StartTime -
                                    ((SRINT64)intsec1970 * (SRINT64)NSPERSEC) );
        
	ttTime      = (time_t)intsec1970;
	microsecond = (int)(nanosec1970 * NSPERUS);

        ptm = gmtime( &ttTime );

        seedlinktime.year   =  ptm->tm_year + 1900;
        seedlinktime.yday   =  ptm->tm_yday + 1;
        seedlinktime.hour   =  ptm->tm_hour;
        seedlinktime.minute =  ptm->tm_min;
        seedlinktime.second =  ptm->tm_sec;
	seedlinktime.usec   =  microsecond;


        if ( PP.Verbosity >= 6 ) {
                sprintf( MsgBuffer, "Ready to send %d pts for ymd %04d/%02d/%02d %02d:%02d:%02d.%06d (intsec1970 %d)\n",
                         PP.NumReadySamples,
                         seedlinktime.year, (ptm->tm_mon+1), ptm->tm_mday,
                         seedlinktime.hour, seedlinktime.minute, seedlinktime.second,
                         seedlinktime.usec, intsec1970 );
                LogMessage( MsgBuffer, 6, 0 );
                }

	
        // Loop around sending raw data to SeisComp.

        for ( ichan = 0 ; ichan < MAX_SC_CHAN ; ichan++ ) {

                if ( PP.Verbosity >= 6 ) {
                        sprintf( MsgBuffer,
                                 "Checking enable for trace %2d for station %s Channel %s\n",
                                 ichan, PP.StationName, PP.ChannelName[ichan] );
                        LogMessage( MsgBuffer, 6, 0 );
                        }

                if ( PP.ChannelEnable[ichan] == 1 ) {

                        if ( FirstTimeMsg < PP.NumChannelEnabled ) {
                                FirstTimeMsg++;				
                                sprintf( MsgBuffer,
                                         "Sending enabled trace %d for station %s Channel %s, nsamples %d\n",
                                         ichan, PP.StationName, PP.ChannelName[ichan], PP.NumReadySamples );
                                LogMessage( MsgBuffer, 2, 0 );
                                }

			r = send_raw3( PP.StationName, PP.ChannelName[ichan],
				       &seedlinktime, 0, TimeQuality,                            // time 
				       (int32_t*)(PP.TraceBuffer[ichan]), PP.NumReadySamples );  // data

			if ( r == -1 ) {
				sprintf( MsgBuffer,
					 "Bin2Sc (sr2sc_plugin) call send_raw3 return value is %d\n", r );
				LogMessage( MsgBuffer, 5, 0 );
				CleanExit( 0 );

                                }
                        
                        } // end if ChannelEnable == 1

		} // end for ichan < MAX_SC_CHAN

        PP.NumReadySamples = 0;
           
        LogMessage( "SendTraces ending ...\n", 6, 0 );

        return;
}


//------------------------------------------------------------------------------
// ROUTINE: ScCheckTerminate
// PURPOSE: Check to see if the Seiscomp system is requesting a termination
//          and exit if it is.
//------------------------------------------------------------------------------
void ScCheckTerminate( void ) {

	LogMessage( "ScCheckTerminate starting ...\n", 7, 0 );

        // Check for Seiscomp terminate request.

        if ( PP.fStopExecution == 1 ) {
		LogMessage( "ScCheckTerminate found terminate request ...\n", 1, 0 );
		CleanExit( 0 );
		}

        LogMessage( "ScCheckTerminate ending ...\n", 7, 0 );
        return;
}


//------------------------------------------------------------------------------
// ROUTINE: CleanExit
// PURPOSE: Do all tasks needed for a clean exit.  These include closing the
//          input pipe, killing the pipeline chain, and closing report files.
//------------------------------------------------------------------------------
void CleanExit( int ExitCode ) {

        LogMessage( "CleanExit starting ...\n", 3, 0 );


	// Close the input pipe.

	CustomCloseInputPipe( PP.hInputPipe );



	// Call batch file to kill the SR to SC pipeline chain.

	sprintf( PipelineCmd, "%s/Sr2ScPipelineKill", SeedlinkDir );
	RunPipelineCmd( PipelineCmd );



        // Clean exit for Seiscomp.

        
        // Notify user.

        LogMessage( "Sr2Sc: Termination requested; exiting!\n", 0, 0 );
	fflush( stdout );



	// Close custom report and debugging file.

	PP.CustomReportReady = 0;
	CustomReportClose();
	if ( fp )  fclose( fp );



        // Bye bye.

	exit( ExitCode );
}


//------------------------------------------------------------------------------
// ROUTINE: SignalInit
// PURPOSE: Initialize catching of TERM and KILL signal.
//------------------------------------------------------------------------------
void SignalInit( void ) {

	struct sigaction newa, olda;

	LogMessage( "SignalInit starting ...\n", 2, 0 );
	
	// Define function called when receiving such signals.

	newa.sa_handler = SignalHandle;
	newa.sa_flags   = SA_NOMASK;            // see 'man sigaction'

	sigaction( SIGTERM, &newa, &olda );     // e.g kill
	sigaction( SIGINT, &newa, &olda );      // e.g CTRL-C

	LogMessage( "SignalInit ending ...\n", 2, 0 );
}

//------------------------------------------------------------------------------
// ROUTINE: SignalHandle
// PURPOSE: Called to request termination when TERM or KILL signal is received.
//------------------------------------------------------------------------------
void SignalHandle( int x ) {

	char MsgBuffer[MAXSTRBIG];

        if ( x == SIGTERM )
                sprintf( MsgBuffer, "SignalHandle starting for a terminate signal (#%d) ...\n", x );
        else if ( x == SIGKILL )
                sprintf( MsgBuffer, "SignalHandle starting for a kill signal (#%d) ...\n", x );
        else
                sprintf( MsgBuffer, "SignalHandle starting for signal %d ...\n", x );
	LogMessage( MsgBuffer, 2, 0 );

	PP.fStopExecution = 1;
	CleanExit( 0 );

	LogMessage( "SignalHandle ending ...\n", 2, 0 );
}


//------------------------------------------------------------------------------
// ROUTINE: LogMessage
// PURPOSE: If specified level is exceeded, the message is logged to the 
//          SeisComp log, the SR report system, and the debug report file.  
//          Then, if an error code was passed, exit.
//------------------------------------------------------------------------------
void LogMessage( char *msg, int level, int err ) {

	struct tm *ptm;
	time_t now;
	char FullMsg[MAXSTR];


	// Quick exit if log level test not met.

	if ( PP.Verbosity < level )
		return;


       // Get current system time.

	time( &now );
	ptm = gmtime( &now );


	// Prepare full message.

	sprintf( FullMsg, "%04d/%02d/%02d %02d:%02d:%02d [sr2sc]: %s",
		  (ptm->tm_year + 1900), ptm->tm_mon, ptm->tm_mday,
		  ptm->tm_hour, ptm->tm_min, ptm->tm_sec, msg );
	


	// Send message to stdout which is captured by SeisComp log.

	printf( "%s", FullMsg );



	// Send message to the SR report system.

	if ( PP.CustomReportReady )
		CustomReportPrint( FullMsg );


	
	// Send message to debug report file.
	
	if ( fp ) {
		fprintf( fp, "%s", FullMsg );
		fflush( fp );
		}



	// Exit on error.

	if ( err )
		CleanExit( err );
}

